/**
 * Created by Samuel Gratzl on 01.09.2015.
 */
import * as tslib_1 from "tslib";
import * as d3 from 'd3';
import * as C from 'phovea_core/src/index';
import * as events from 'phovea_core/src/event';
/**
 * normalizes the given coordinates to sum up to one
 * @param arr
 * @returns {any}
 */
function normalize(arr) {
    var sum = arr.reduce(function (a, b) { return a + b; }, 0);
    return arr.map(function (i) { return i / sum; });
}
/**
 * generic version of the CLUE mode, a combination of exploration, authoring, and normalization
 */
var CLUEMode = (function () {
    function CLUEMode(exploration, authoring, presentation) {
        this.coord = normalize([exploration, authoring, presentation]);
    }
    Object.defineProperty(CLUEMode.prototype, "exploration", {
        get: function () {
            return this.coord[0];
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CLUEMode.prototype, "authoring", {
        get: function () {
            return this.coord[1];
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CLUEMode.prototype, "presentation", {
        get: function () {
            return this.coord[2];
        },
        enumerable: true,
        configurable: true
    });
    CLUEMode.prototype.value = function (index) {
        if (typeof index === 'number') {
            return this.coord[index];
        }
        else if (typeof index === 'string') {
            var lookup = { e: this.coord[0], a: this.coord[1], p: this.coord[2] };
            return lookup[index.charAt(0).toLowerCase()];
        }
        return null;
    };
    Object.defineProperty(CLUEMode.prototype, "isAtomic", {
        /**
         * whether this mode is extreme, i.e., in one corner of the triangle
         * @returns {boolean}
         */
        get: function () {
            return this.exploration === 1.0 || this.authoring === 1.0 || this.presentation === 1.0;
        },
        enumerable: true,
        configurable: true
    });
    CLUEMode.prototype.toString = function () {
        if (this.exploration === 1) {
            return 'E';
        }
        if (this.authoring === 1) {
            return 'A';
        }
        if (this.presentation === 1) {
            return 'P';
        }
        return '(' + this.coord.map(function (s) { return d3.round(s, 3).toString(); }).join('-') + ')';
    };
    return CLUEMode;
}());
export { CLUEMode };
/**
 * mode factory by the given components
 * @param exploration
 * @param authoring
 * @param presentation
 * @returns {CLUEMode}
 */
function mode(exploration, authoring, presentation) {
    return new CLUEMode(exploration, authoring, presentation);
}
/**
 * shortcuts for the atomic modes
 * @type {{Exploration: CLUEMode, Authoring: CLUEMode, Presentation: CLUEMode}}
 */
export var modes = {
    Exploration: mode(1, 0, 0),
    Authoring: mode(0, 1, 0),
    Presentation: mode(0, 0, 1)
};
function fromString(s) {
    if (s === 'P') {
        return modes.Presentation;
    }
    else if (s === 'A') {
        return modes.Authoring;
    }
    else if (s === 'E') {
        return modes.Exploration;
    }
    var coords = s.slice(1, s.length - 1).split('-').map(parseFloat);
    return new CLUEMode(coords[0], coords[1], coords[2]);
}
/**
 * returns the default mode either stored in the hash or by default exploration
 */
function defaultMode() {
    return fromString(C.hash.getProp('clue', 'E'));
}
/**
 * wrapper containing the current mode
 */
var ModeWrapper = (function (_super) {
    tslib_1.__extends(ModeWrapper, _super);
    function ModeWrapper() {
        var _this = _super.call(this) || this;
        _this._mode = defaultMode();
        events.fire('clue.modeChanged', _this._mode, _this._mode);
        return _this;
    }
    Object.defineProperty(ModeWrapper.prototype, "mode", {
        get: function () {
            return this._mode;
        },
        set: function (value) {
            if (this._mode === value) {
                return;
            }
            if (value.isAtomic) {
                //use the real atomic one for a shared instance
                value = fromString(value.toString());
            }
            var bak = this._mode;
            this._mode = value;
            //store in hash
            C.hash.setProp('clue', value.toString());
            this.fire('modeChanged', value, bak);
            events.fire('clue.modeChanged', value, bak);
        },
        enumerable: true,
        configurable: true
    });
    return ModeWrapper;
}(events.EventHandler));
var _instance = new ModeWrapper();
export var on = ModeWrapper.prototype.on.bind(_instance);
export var off = ModeWrapper.prototype.off.bind(_instance);
/**
 * returns the current mode
 * @returns {CLUEMode}
 */
export function getMode() {
    return _instance.mode;
}
/**
 * set the mode
 * @param value
 */
export function setMode(value) {
    _instance.mode = value;
}
/**
 * utility to select the mode using three buttons to the atomic versions using bootstrap buttons
 */
var ButtonModeSelector = (function () {
    function ButtonModeSelector(parent, options) {
        if (options === void 0) { options = {}; }
        var _this = this;
        this.options = {
            /**
             * button size, i.e. the class btn-{size] will be added
             */
            size: 'xs'
        };
        C.mixin(this.options, options);
        this.$node = this.build(d3.select(parent));
        var listener = function (event, newMode) {
            _this.$node.selectAll('label').classed('active', function (d) { return d === newMode; }).select('input').property('checked', function (d) { return d === newMode; });
        };
        _instance.on('modeChanged', listener);
        C.onDOMNodeRemoved(this.$node.node(), function () {
            _instance.off('modeChanged', listener);
        });
    }
    ButtonModeSelector.prototype.build = function ($parent) {
        var _this = this;
        var $root = $parent.append('div').classed('clue_buttonmodeselector', true).classed('btn-group', true).attr('data-toggle', 'buttons');
        var $modes = $root.selectAll('label').data([modes.Exploration, modes.Authoring, modes.Presentation]);
        $modes.enter().append('label')
            .attr('class', function (d) { return 'btn btn-' + _this.options.size + ' clue-' + d.toString(); })
            .classed('active', function (d) { return d === getMode(); })
            .html(function (d, i) {
            var label = ['Exploration', 'Authoring', 'Presentation'][i];
            return "<input type=\"radio\" name=\"clue_mode\" autocomplete=\"off\" value=\"" + d.toString() + "\" " + (d === getMode() ? 'checked="checked"' : '') + "> " + label;
        }).on('click', function (d) {
            setMode(d);
        });
        return $root;
    };
    return ButtonModeSelector;
}());
export { ButtonModeSelector };
/**
 * mode selector based on three sliders for each dimensions that are synced
 */
var SliderModeSelector = (function () {
    function SliderModeSelector(parent, options) {
        if (options === void 0) { options = {}; }
        var _this = this;
        this.options = {};
        C.mixin(this.options, options);
        this.$node = d3.select(parent).append('div').classed('clue_modeselector', true).datum(this);
        this.build(this.$node);
        var listener = function (event, newMode) {
            _this.$node.select('label.clue-E input').property('value', Math.round(newMode.exploration * 100));
            _this.$node.select('label.clue-A input').property('value', Math.round(newMode.authoring * 100));
            _this.$node.select('label.clue-P input').property('value', Math.round(newMode.presentation * 100));
        };
        _instance.on('modeChanged', listener);
        C.onDOMNodeRemoved(this.$node.node(), function () {
            _instance.off('modeChanged', listener);
        });
    }
    SliderModeSelector.prototype.build = function ($parent) {
        var $root = $parent.append('div').classed('clue_slidermodeselector', true);
        var $modes = $root.selectAll('label').data([modes.Exploration, modes.Authoring, modes.Presentation]);
        function normalize(eap, drivenBy) {
            var base = eap[drivenBy];
            eap[drivenBy] = 0;
            var factor = (1 - base) / d3.sum(eap);
            eap = eap.map(function (v) { return v * factor; });
            eap[drivenBy] = base;
            return eap;
        }
        function updateMode(drivenBy) {
            if (drivenBy === void 0) { drivenBy = -1; }
            var e = parseFloat($modes.select('label.clue-E input').property('value')) / 100;
            var a = parseFloat($modes.select('label.clue-A input').property('value')) / 100;
            var p = parseFloat($modes.select('label.clue-P input').property('value')) / 100;
            if (drivenBy >= 0) {
                _a = normalize([e, a, p], drivenBy), e = _a[0], a = _a[1], p = _a[2];
            }
            setMode(mode(e, a, p));
            var _a;
        }
        $modes.enter().append('label')
            .attr('class', function (d) { return 'clue-' + d.toString(); })
            .text(function (d, i) { return ['Exploration', 'Authoring', 'Presentation'][i]; })
            .append('input')
            .attr({
            type: 'range',
            min: 0,
            max: 100,
            value: function (d, i) { return getMode().value(i) * 100; }
        }).on('input', function (d, i) {
            updateMode(i);
        });
        return $root;
    };
    return SliderModeSelector;
}());
export { SliderModeSelector };
/**
 * mode selector based on a triangle
 */
var TriangleModeSelector = (function () {
    function TriangleModeSelector(parent, options) {
        if (options === void 0) { options = {}; }
        var _this = this;
        this.options = {
            /**
             * height of the triangle
             */
            height: 15,
            /**
             * offset bounds
             */
            offset: 5
        };
        this.e = [0, 30];
        this.a = [30, 0];
        this.p = [60, 30];
        C.mixin(this.options, options);
        this.e[1] = this.a[0] = this.p[1] = this.options.height;
        this.p[0] = this.options.height * 2;
        this.$node = d3.select(parent).append('div').classed('clue_trianglemodeselector', true).datum(this);
        this.build(this.$node);
        var listener = function (event, newMode) {
            var c = _this.toCoordinates(newMode);
            _this.$node.select('circle.point').attr({
                cx: c[0],
                cy: c[1]
            });
        };
        _instance.on('modeChanged', listener);
        C.onDOMNodeRemoved(this.$node.node(), function () {
            _instance.off('modeChanged', listener);
        });
    }
    TriangleModeSelector.prototype.toCoordinates = function (m) {
        var x = m.exploration * this.e[0] + m.authoring * this.a[0] + m.presentation * this.p[0];
        var y = m.exploration * this.e[1] + m.authoring * this.a[1] + m.presentation * this.p[1];
        return [x, y];
    };
    TriangleModeSelector.prototype.fromCoordinates = function (x, y) {
        //https://en.wikipedia.org/wiki/Barycentric_coordinate_system
        var x1 = this.e[0], x2 = this.a[0], x3 = this.p[0], y1 = this.e[1], y2 = this.a[1], y3 = this.p[1];
        var e = Math.max(0, Math.min(1, ((y2 - y3) * (x - x3) + (x3 - x2) * (y - y3)) / ((y2 - y3) * (x1 - x3) + (x3 - x2) * (y1 - y3))));
        var a = Math.max(0, Math.min(1, ((y3 - y3) * (x - x3) + (x1 - x3) * (y - y3)) / ((y2 - y3) * (x1 - x3) + (x3 - x2) * (y1 - y3))));
        var s = e + a;
        if (s > 1) {
            e /= s;
            a /= s;
        }
        var p = 1 - e - a;
        return mode(e, a, p);
    };
    TriangleModeSelector.prototype.build = function ($parent) {
        var _this = this;
        var $root = $parent.append('svg').classed('clue_trianglemodeselector', true).attr({
            width: this.p[0] + this.options.offset,
            height: this.p[1] + this.options.offset
        });
        var that = this;
        var $g = $root.append('g').attr('transform', "translate(" + this.options.offset / 2 + "," + this.options.offset / 2 + ")");
        $g.append('path').attr('d', d3.svg.line().interpolate('linear-closed')([this.e, this.a, this.p])).on('click', function () {
            var xy = d3.mouse(this);
            var m = that.fromCoordinates(xy[0], xy[1]);
            setMode(m);
        });
        var xy = this.toCoordinates(getMode());
        $g.append('circle').classed('point', true).attr({
            cx: xy[0],
            cy: xy[1],
            r: 2
        }).call(d3.behavior.drag().on('drag', function () {
            var m = _this.fromCoordinates(d3.event.x, d3.event.y);
            setMode(m);
        }));
        return $root;
    };
    return TriangleModeSelector;
}());
export { TriangleModeSelector };
/**
 * alias for `createTriangle`
 * @param parent the parent dom element
 * @param options
 * @returns {TriangleModeSelector}
 */
export function create(parent, options) {
    if (options === void 0) { options = {}; }
    return createTriangle(parent, options);
}
export function createTriangle(parent, options) {
    if (options === void 0) { options = {}; }
    return new TriangleModeSelector(parent, options);
}
export function createButton(parent, options) {
    if (options === void 0) { options = {}; }
    return new ButtonModeSelector(parent, options);
}
export function createSlider(parent, options) {
    if (options === void 0) { options = {}; }
    return new SliderModeSelector(parent, options);
}
//# sourceMappingURL=mode.js.map